# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from lndmanage.grpc_compiled import signer_pb2 as signer__pb2
from lndmanage.grpc_compiled import walletkit_pb2 as walletkit__pb2


class WalletKitStub(object):
    """WalletKit is a service that gives access to the core functionalities of the
    daemon's wallet.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListUnspent = channel.unary_unary(
                '/walletrpc.WalletKit/ListUnspent',
                request_serializer=walletkit__pb2.ListUnspentRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ListUnspentResponse.FromString,
                )
        self.LeaseOutput = channel.unary_unary(
                '/walletrpc.WalletKit/LeaseOutput',
                request_serializer=walletkit__pb2.LeaseOutputRequest.SerializeToString,
                response_deserializer=walletkit__pb2.LeaseOutputResponse.FromString,
                )
        self.ReleaseOutput = channel.unary_unary(
                '/walletrpc.WalletKit/ReleaseOutput',
                request_serializer=walletkit__pb2.ReleaseOutputRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ReleaseOutputResponse.FromString,
                )
        self.ListLeases = channel.unary_unary(
                '/walletrpc.WalletKit/ListLeases',
                request_serializer=walletkit__pb2.ListLeasesRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ListLeasesResponse.FromString,
                )
        self.DeriveNextKey = channel.unary_unary(
                '/walletrpc.WalletKit/DeriveNextKey',
                request_serializer=walletkit__pb2.KeyReq.SerializeToString,
                response_deserializer=signer__pb2.KeyDescriptor.FromString,
                )
        self.DeriveKey = channel.unary_unary(
                '/walletrpc.WalletKit/DeriveKey',
                request_serializer=signer__pb2.KeyLocator.SerializeToString,
                response_deserializer=signer__pb2.KeyDescriptor.FromString,
                )
        self.NextAddr = channel.unary_unary(
                '/walletrpc.WalletKit/NextAddr',
                request_serializer=walletkit__pb2.AddrRequest.SerializeToString,
                response_deserializer=walletkit__pb2.AddrResponse.FromString,
                )
        self.ListAccounts = channel.unary_unary(
                '/walletrpc.WalletKit/ListAccounts',
                request_serializer=walletkit__pb2.ListAccountsRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ListAccountsResponse.FromString,
                )
        self.ImportAccount = channel.unary_unary(
                '/walletrpc.WalletKit/ImportAccount',
                request_serializer=walletkit__pb2.ImportAccountRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ImportAccountResponse.FromString,
                )
        self.ImportPublicKey = channel.unary_unary(
                '/walletrpc.WalletKit/ImportPublicKey',
                request_serializer=walletkit__pb2.ImportPublicKeyRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ImportPublicKeyResponse.FromString,
                )
        self.PublishTransaction = channel.unary_unary(
                '/walletrpc.WalletKit/PublishTransaction',
                request_serializer=walletkit__pb2.Transaction.SerializeToString,
                response_deserializer=walletkit__pb2.PublishResponse.FromString,
                )
        self.SendOutputs = channel.unary_unary(
                '/walletrpc.WalletKit/SendOutputs',
                request_serializer=walletkit__pb2.SendOutputsRequest.SerializeToString,
                response_deserializer=walletkit__pb2.SendOutputsResponse.FromString,
                )
        self.EstimateFee = channel.unary_unary(
                '/walletrpc.WalletKit/EstimateFee',
                request_serializer=walletkit__pb2.EstimateFeeRequest.SerializeToString,
                response_deserializer=walletkit__pb2.EstimateFeeResponse.FromString,
                )
        self.PendingSweeps = channel.unary_unary(
                '/walletrpc.WalletKit/PendingSweeps',
                request_serializer=walletkit__pb2.PendingSweepsRequest.SerializeToString,
                response_deserializer=walletkit__pb2.PendingSweepsResponse.FromString,
                )
        self.BumpFee = channel.unary_unary(
                '/walletrpc.WalletKit/BumpFee',
                request_serializer=walletkit__pb2.BumpFeeRequest.SerializeToString,
                response_deserializer=walletkit__pb2.BumpFeeResponse.FromString,
                )
        self.ListSweeps = channel.unary_unary(
                '/walletrpc.WalletKit/ListSweeps',
                request_serializer=walletkit__pb2.ListSweepsRequest.SerializeToString,
                response_deserializer=walletkit__pb2.ListSweepsResponse.FromString,
                )
        self.LabelTransaction = channel.unary_unary(
                '/walletrpc.WalletKit/LabelTransaction',
                request_serializer=walletkit__pb2.LabelTransactionRequest.SerializeToString,
                response_deserializer=walletkit__pb2.LabelTransactionResponse.FromString,
                )
        self.FundPsbt = channel.unary_unary(
                '/walletrpc.WalletKit/FundPsbt',
                request_serializer=walletkit__pb2.FundPsbtRequest.SerializeToString,
                response_deserializer=walletkit__pb2.FundPsbtResponse.FromString,
                )
        self.SignPsbt = channel.unary_unary(
                '/walletrpc.WalletKit/SignPsbt',
                request_serializer=walletkit__pb2.SignPsbtRequest.SerializeToString,
                response_deserializer=walletkit__pb2.SignPsbtResponse.FromString,
                )
        self.FinalizePsbt = channel.unary_unary(
                '/walletrpc.WalletKit/FinalizePsbt',
                request_serializer=walletkit__pb2.FinalizePsbtRequest.SerializeToString,
                response_deserializer=walletkit__pb2.FinalizePsbtResponse.FromString,
                )


class WalletKitServicer(object):
    """WalletKit is a service that gives access to the core functionalities of the
    daemon's wallet.
    """

    def ListUnspent(self, request, context):
        """
        ListUnspent returns a list of all utxos spendable by the wallet with a
        number of confirmations between the specified minimum and maximum.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LeaseOutput(self, request, context):
        """
        LeaseOutput locks an output to the given ID, preventing it from being
        available for any future coin selection attempts. The absolute time of the
        lock's expiration is returned. The expiration of the lock can be extended by
        successive invocations of this RPC. Outputs can be unlocked before their
        expiration through `ReleaseOutput`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReleaseOutput(self, request, context):
        """
        ReleaseOutput unlocks an output, allowing it to be available for coin
        selection if it remains unspent. The ID should match the one used to
        originally lock the output.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListLeases(self, request, context):
        """
        ListLeases lists all currently locked utxos.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeriveNextKey(self, request, context):
        """
        DeriveNextKey attempts to derive the *next* key within the key family
        (account in BIP43) specified. This method should return the next external
        child within this branch.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeriveKey(self, request, context):
        """
        DeriveKey attempts to derive an arbitrary key specified by the passed
        KeyLocator.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def NextAddr(self, request, context):
        """
        NextAddr returns the next unused address within the wallet.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAccounts(self, request, context):
        """
        ListAccounts retrieves all accounts belonging to the wallet by default. A
        name and key scope filter can be provided to filter through all of the
        wallet accounts and return only those matching.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImportAccount(self, request, context):
        """
        ImportAccount imports an account backed by an account extended public key.
        The master key fingerprint denotes the fingerprint of the root key
        corresponding to the account public key (also known as the key with
        derivation path m/). This may be required by some hardware wallets for
        proper identification and signing.

        The address type can usually be inferred from the key's version, but may be
        required for certain keys to map them into the proper scope.

        For BIP-0044 keys, an address type must be specified as we intend to not
        support importing BIP-0044 keys into the wallet using the legacy
        pay-to-pubkey-hash (P2PKH) scheme. A nested witness address type will force
        the standard BIP-0049 derivation scheme, while a witness address type will
        force the standard BIP-0084 derivation scheme.

        For BIP-0049 keys, an address type must also be specified to make a
        distinction between the standard BIP-0049 address schema (nested witness
        pubkeys everywhere) and our own BIP-0049Plus address schema (nested pubkeys
        externally, witness pubkeys internally).

        NOTE: Events (deposits/spends) for keys derived from an account will only be
        detected by lnd if they happen after the import. Rescans to detect past
        events will be supported later on.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImportPublicKey(self, request, context):
        """
        ImportPublicKey imports a public key as watch-only into the wallet.

        NOTE: Events (deposits/spends) for a key will only be detected by lnd if
        they happen after the import. Rescans to detect past events will be
        supported later on.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PublishTransaction(self, request, context):
        """
        PublishTransaction attempts to publish the passed transaction to the
        network. Once this returns without an error, the wallet will continually
        attempt to re-broadcast the transaction on start up, until it enters the
        chain.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendOutputs(self, request, context):
        """
        SendOutputs is similar to the existing sendmany call in Bitcoind, and
        allows the caller to create a transaction that sends to several outputs at
        once. This is ideal when wanting to batch create a set of transactions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EstimateFee(self, request, context):
        """
        EstimateFee attempts to query the internal fee estimator of the wallet to
        determine the fee (in sat/kw) to attach to a transaction in order to
        achieve the confirmation target.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PendingSweeps(self, request, context):
        """
        PendingSweeps returns lists of on-chain outputs that lnd is currently
        attempting to sweep within its central batching engine. Outputs with similar
        fee rates are batched together in order to sweep them within a single
        transaction.

        NOTE: Some of the fields within PendingSweepsRequest are not guaranteed to
        remain supported. This is an advanced API that depends on the internals of
        the UtxoSweeper, so things may change.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BumpFee(self, request, context):
        """
        BumpFee bumps the fee of an arbitrary input within a transaction. This RPC
        takes a different approach than bitcoind's bumpfee command. lnd has a
        central batching engine in which inputs with similar fee rates are batched
        together to save on transaction fees. Due to this, we cannot rely on
        bumping the fee on a specific transaction, since transactions can change at
        any point with the addition of new inputs. The list of inputs that
        currently exist within lnd's central batching engine can be retrieved
        through the PendingSweeps RPC.

        When bumping the fee of an input that currently exists within lnd's central
        batching engine, a higher fee transaction will be created that replaces the
        lower fee transaction through the Replace-By-Fee (RBF) policy. If it

        This RPC also serves useful when wanting to perform a Child-Pays-For-Parent
        (CPFP), where the child transaction pays for its parent's fee. This can be
        done by specifying an outpoint within the low fee transaction that is under
        the control of the wallet.

        The fee preference can be expressed either as a specific fee rate or a delta
        of blocks in which the output should be swept on-chain within. If a fee
        preference is not explicitly specified, then an error is returned.

        Note that this RPC currently doesn't perform any validation checks on the
        fee preference being provided. For now, the responsibility of ensuring that
        the new fee preference is sufficient is delegated to the user.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListSweeps(self, request, context):
        """
        ListSweeps returns a list of the sweep transactions our node has produced.
        Note that these sweeps may not be confirmed yet, as we record sweeps on
        broadcast, not confirmation.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LabelTransaction(self, request, context):
        """
        LabelTransaction adds a label to a transaction. If the transaction already
        has a label the call will fail unless the overwrite bool is set. This will
        overwrite the exiting transaction label. Labels must not be empty, and
        cannot exceed 500 characters.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FundPsbt(self, request, context):
        """
        FundPsbt creates a fully populated PSBT that contains enough inputs to fund
        the outputs specified in the template. There are two ways of specifying a
        template: Either by passing in a PSBT with at least one output declared or
        by passing in a raw TxTemplate message.

        If there are no inputs specified in the template, coin selection is
        performed automatically. If the template does contain any inputs, it is
        assumed that full coin selection happened externally and no additional
        inputs are added. If the specified inputs aren't enough to fund the outputs
        with the given fee rate, an error is returned.

        After either selecting or verifying the inputs, all input UTXOs are locked
        with an internal app ID.

        NOTE: If this method returns without an error, it is the caller's
        responsibility to either spend the locked UTXOs (by finalizing and then
        publishing the transaction) or to unlock/release the locked UTXOs in case of
        an error on the caller's side.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SignPsbt(self, request, context):
        """
        SignPsbt expects a partial transaction with all inputs and outputs fully
        declared and tries to sign all unsigned inputs that have all required fields
        (UTXO information, BIP32 derivation information, witness or sig scripts)
        set.
        If no error is returned, the PSBT is ready to be given to the next signer or
        to be finalized if lnd was the last signer.

        NOTE: This RPC only signs inputs (and only those it can sign), it does not
        perform any other tasks (such as coin selection, UTXO locking or
        input/output/fee value validation, PSBT finalization). Any input that is
        incomplete will be skipped.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FinalizePsbt(self, request, context):
        """
        FinalizePsbt expects a partial transaction with all inputs and outputs fully
        declared and tries to sign all inputs that belong to the wallet. Lnd must be
        the last signer of the transaction. That means, if there are any unsigned
        non-witness inputs or inputs without UTXO information attached or inputs
        without witness data that do not belong to lnd's wallet, this method will
        fail. If no error is returned, the PSBT is ready to be extracted and the
        final TX within to be broadcast.

        NOTE: This method does NOT publish the transaction once finalized. It is the
        caller's responsibility to either publish the transaction on success or
        unlock/release any locked UTXOs in case of an error in this method.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WalletKitServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListUnspent': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUnspent,
                    request_deserializer=walletkit__pb2.ListUnspentRequest.FromString,
                    response_serializer=walletkit__pb2.ListUnspentResponse.SerializeToString,
            ),
            'LeaseOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.LeaseOutput,
                    request_deserializer=walletkit__pb2.LeaseOutputRequest.FromString,
                    response_serializer=walletkit__pb2.LeaseOutputResponse.SerializeToString,
            ),
            'ReleaseOutput': grpc.unary_unary_rpc_method_handler(
                    servicer.ReleaseOutput,
                    request_deserializer=walletkit__pb2.ReleaseOutputRequest.FromString,
                    response_serializer=walletkit__pb2.ReleaseOutputResponse.SerializeToString,
            ),
            'ListLeases': grpc.unary_unary_rpc_method_handler(
                    servicer.ListLeases,
                    request_deserializer=walletkit__pb2.ListLeasesRequest.FromString,
                    response_serializer=walletkit__pb2.ListLeasesResponse.SerializeToString,
            ),
            'DeriveNextKey': grpc.unary_unary_rpc_method_handler(
                    servicer.DeriveNextKey,
                    request_deserializer=walletkit__pb2.KeyReq.FromString,
                    response_serializer=signer__pb2.KeyDescriptor.SerializeToString,
            ),
            'DeriveKey': grpc.unary_unary_rpc_method_handler(
                    servicer.DeriveKey,
                    request_deserializer=signer__pb2.KeyLocator.FromString,
                    response_serializer=signer__pb2.KeyDescriptor.SerializeToString,
            ),
            'NextAddr': grpc.unary_unary_rpc_method_handler(
                    servicer.NextAddr,
                    request_deserializer=walletkit__pb2.AddrRequest.FromString,
                    response_serializer=walletkit__pb2.AddrResponse.SerializeToString,
            ),
            'ListAccounts': grpc.unary_unary_rpc_method_handler(
                    servicer.ListAccounts,
                    request_deserializer=walletkit__pb2.ListAccountsRequest.FromString,
                    response_serializer=walletkit__pb2.ListAccountsResponse.SerializeToString,
            ),
            'ImportAccount': grpc.unary_unary_rpc_method_handler(
                    servicer.ImportAccount,
                    request_deserializer=walletkit__pb2.ImportAccountRequest.FromString,
                    response_serializer=walletkit__pb2.ImportAccountResponse.SerializeToString,
            ),
            'ImportPublicKey': grpc.unary_unary_rpc_method_handler(
                    servicer.ImportPublicKey,
                    request_deserializer=walletkit__pb2.ImportPublicKeyRequest.FromString,
                    response_serializer=walletkit__pb2.ImportPublicKeyResponse.SerializeToString,
            ),
            'PublishTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.PublishTransaction,
                    request_deserializer=walletkit__pb2.Transaction.FromString,
                    response_serializer=walletkit__pb2.PublishResponse.SerializeToString,
            ),
            'SendOutputs': grpc.unary_unary_rpc_method_handler(
                    servicer.SendOutputs,
                    request_deserializer=walletkit__pb2.SendOutputsRequest.FromString,
                    response_serializer=walletkit__pb2.SendOutputsResponse.SerializeToString,
            ),
            'EstimateFee': grpc.unary_unary_rpc_method_handler(
                    servicer.EstimateFee,
                    request_deserializer=walletkit__pb2.EstimateFeeRequest.FromString,
                    response_serializer=walletkit__pb2.EstimateFeeResponse.SerializeToString,
            ),
            'PendingSweeps': grpc.unary_unary_rpc_method_handler(
                    servicer.PendingSweeps,
                    request_deserializer=walletkit__pb2.PendingSweepsRequest.FromString,
                    response_serializer=walletkit__pb2.PendingSweepsResponse.SerializeToString,
            ),
            'BumpFee': grpc.unary_unary_rpc_method_handler(
                    servicer.BumpFee,
                    request_deserializer=walletkit__pb2.BumpFeeRequest.FromString,
                    response_serializer=walletkit__pb2.BumpFeeResponse.SerializeToString,
            ),
            'ListSweeps': grpc.unary_unary_rpc_method_handler(
                    servicer.ListSweeps,
                    request_deserializer=walletkit__pb2.ListSweepsRequest.FromString,
                    response_serializer=walletkit__pb2.ListSweepsResponse.SerializeToString,
            ),
            'LabelTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.LabelTransaction,
                    request_deserializer=walletkit__pb2.LabelTransactionRequest.FromString,
                    response_serializer=walletkit__pb2.LabelTransactionResponse.SerializeToString,
            ),
            'FundPsbt': grpc.unary_unary_rpc_method_handler(
                    servicer.FundPsbt,
                    request_deserializer=walletkit__pb2.FundPsbtRequest.FromString,
                    response_serializer=walletkit__pb2.FundPsbtResponse.SerializeToString,
            ),
            'SignPsbt': grpc.unary_unary_rpc_method_handler(
                    servicer.SignPsbt,
                    request_deserializer=walletkit__pb2.SignPsbtRequest.FromString,
                    response_serializer=walletkit__pb2.SignPsbtResponse.SerializeToString,
            ),
            'FinalizePsbt': grpc.unary_unary_rpc_method_handler(
                    servicer.FinalizePsbt,
                    request_deserializer=walletkit__pb2.FinalizePsbtRequest.FromString,
                    response_serializer=walletkit__pb2.FinalizePsbtResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'walletrpc.WalletKit', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class WalletKit(object):
    """WalletKit is a service that gives access to the core functionalities of the
    daemon's wallet.
    """

    @staticmethod
    def ListUnspent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ListUnspent',
            walletkit__pb2.ListUnspentRequest.SerializeToString,
            walletkit__pb2.ListUnspentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LeaseOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/LeaseOutput',
            walletkit__pb2.LeaseOutputRequest.SerializeToString,
            walletkit__pb2.LeaseOutputResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReleaseOutput(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ReleaseOutput',
            walletkit__pb2.ReleaseOutputRequest.SerializeToString,
            walletkit__pb2.ReleaseOutputResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListLeases(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ListLeases',
            walletkit__pb2.ListLeasesRequest.SerializeToString,
            walletkit__pb2.ListLeasesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeriveNextKey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/DeriveNextKey',
            walletkit__pb2.KeyReq.SerializeToString,
            signer__pb2.KeyDescriptor.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeriveKey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/DeriveKey',
            signer__pb2.KeyLocator.SerializeToString,
            signer__pb2.KeyDescriptor.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def NextAddr(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/NextAddr',
            walletkit__pb2.AddrRequest.SerializeToString,
            walletkit__pb2.AddrResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListAccounts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ListAccounts',
            walletkit__pb2.ListAccountsRequest.SerializeToString,
            walletkit__pb2.ListAccountsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImportAccount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ImportAccount',
            walletkit__pb2.ImportAccountRequest.SerializeToString,
            walletkit__pb2.ImportAccountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImportPublicKey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ImportPublicKey',
            walletkit__pb2.ImportPublicKeyRequest.SerializeToString,
            walletkit__pb2.ImportPublicKeyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PublishTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/PublishTransaction',
            walletkit__pb2.Transaction.SerializeToString,
            walletkit__pb2.PublishResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendOutputs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/SendOutputs',
            walletkit__pb2.SendOutputsRequest.SerializeToString,
            walletkit__pb2.SendOutputsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EstimateFee(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/EstimateFee',
            walletkit__pb2.EstimateFeeRequest.SerializeToString,
            walletkit__pb2.EstimateFeeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PendingSweeps(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/PendingSweeps',
            walletkit__pb2.PendingSweepsRequest.SerializeToString,
            walletkit__pb2.PendingSweepsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BumpFee(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/BumpFee',
            walletkit__pb2.BumpFeeRequest.SerializeToString,
            walletkit__pb2.BumpFeeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListSweeps(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/ListSweeps',
            walletkit__pb2.ListSweepsRequest.SerializeToString,
            walletkit__pb2.ListSweepsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LabelTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/LabelTransaction',
            walletkit__pb2.LabelTransactionRequest.SerializeToString,
            walletkit__pb2.LabelTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FundPsbt(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/FundPsbt',
            walletkit__pb2.FundPsbtRequest.SerializeToString,
            walletkit__pb2.FundPsbtResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SignPsbt(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/SignPsbt',
            walletkit__pb2.SignPsbtRequest.SerializeToString,
            walletkit__pb2.SignPsbtResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FinalizePsbt(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/walletrpc.WalletKit/FinalizePsbt',
            walletkit__pb2.FinalizePsbtRequest.SerializeToString,
            walletkit__pb2.FinalizePsbtResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from unittest import TestCase
from lndmanage.lib.ln_utilities import (
    local_balance_to_unbalancedness,
    unbalancedness_to_local_balance,
)


class LnUtilityTest(TestCase):
    def test_unbalancedness_formula(self):
        self.assertAlmostEqual(
            local_balance_to_unbalancedness(500000, 1000000, 1000, False)[0], 0.0
        )
        self.assertAlmostEqual(
            local_balance_to_unbalancedness(500000, 1000000, 1000, True)[0], -0.002
        )
        self.assertAlmostEqual(
            local_balance_to_unbalancedness(600000, 1000000, 0, False)[0], -0.2
        )

        # test inverse:
        ub = -0.2
        cap = 1000000
        cf = 100
        self.assertAlmostEqual(
            ub,
            local_balance_to_unbalancedness(
                unbalancedness_to_local_balance(ub, cap, 0, False)[0], cap, 0, False
            )[0],
        )
        self.assertAlmostEqual(
            ub,
            local_balance_to_unbalancedness(
                unbalancedness_to_local_balance(ub, cap, cf, True)[0], cap, cf, True
            )[0],
        )



import os
import time
import pickle
from typing import Dict, TYPE_CHECKING

import networkx as nx

from lndmanage.lib.utilities import profiled
from lndmanage.lib.ln_utilities import convert_channel_id_to_short_channel_id
from lndmanage.lib.liquidityhints import LiquidityHintMgr
from lndmanage.lib.rating import ChannelRater
from lndmanage import settings

if TYPE_CHECKING:
    from lndmanage.lib.node import LndNode

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def make_cache_filename(filename: str):
    """Creates the cache directory and gives back the absolute path to it for filename."""
    cache_dir = os.path.join(settings.home_dir, 'cache')
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)
    return os.path.join(cache_dir, filename)


class Network:
    """
    Contains the network graph.

    The graph is received from the LND API or from a cached file,
    which contains the graph younger than `settings.CACHING_RETENTION_MINUTES`.
    """
    node: 'LndNode'
    edges: Dict
    graph: nx.MultiGraph
    liquidity_hints: LiquidityHintMgr

    def __init__(self, node: 'LndNode'):
        self.node = node
        self.load_graph()
        self.load_liquidity_hints()
        self.channel_rater = ChannelRater(self)

    @profiled
    def load_graph(self):
        """
        Checks if networkx and edges dictionary pickles are present. If they are older than
        CACHING_RETENTION_MINUTES, make fresh pickles, else read them from the files.
        """
        cache_edges_filename = make_cache_filename('graph.gpickle')
        cache_graph_filename = make_cache_filename('edges.gpickle')

        try:
            timestamp_graph = os.path.getmtime(cache_graph_filename)
        except FileNotFoundError:
            timestamp_graph = 0  # set very old timestamp

        if timestamp_graph < time.time() - settings.CACHING_RETENTION_MINUTES * 60:  # old graph in file
            logger.info(f"Cached graph is too old. Fetching new one.")
            self.set_graph_and_edges()
            nx.write_gpickle(self.graph, cache_graph_filename)
            with open(cache_edges_filename, 'wb') as file:
                pickle.dump(self.edges, file)
        else:  # recent graph in file
            self.graph = nx.read_gpickle(cache_graph_filename)
            with open(cache_edges_filename, 'rb') as file:
                self.edges = pickle.load(file)
            logger.info(f"> Loaded graph from file: {len(self.graph)} nodes, {len(self.edges)} channels.")

    @profiled
    def load_liquidity_hints(self):
        cache_hints_filename = make_cache_filename('liquidity_hints.gpickle')
        try:
            with open(cache_hints_filename, 'rb') as file:
                self.liquidity_hints = pickle.load(file)
                num_badness_hints = len([f for f in self.liquidity_hints._badness_hints.values() if f])
            logger.info(f"> Loaded liquidty hints: {len(self.liquidity_hints._liquidity_hints)} hints, {num_badness_hints} badness hints.")
        except FileNotFoundError:
            self.liquidity_hints = LiquidityHintMgr(self.node.pub_key)
        except Exception as e:
            logger.exception(e)

    @profiled
    def save_liquidty_hints(self):
        cache_hints_filename = make_cache_filename('liquidity_hints.gpickle')
        with open(cache_hints_filename, 'wb') as file:
            pickle.dump(self.liquidity_hints, file)

    @profiled
    def set_graph_and_edges(self):
        """
        Reads in the networkx graph and edges dictionary.

        :return: nx graph and edges dict
        """
        self.edges = {}
        self.graph = nx.MultiGraph()
        raw_graph = self.node.get_raw_network_graph()

        for n in raw_graph.nodes:
            if n.addresses:
                # TODO: handle also ipv6 and onion addresses
                address = n.addresses[0].addr
                if 'onion' in address or '[' in address:
                    address = ''
            else:
                address = ''

            self.graph.add_node(
                n.pub_key,
                alias=n.alias.encode("ascii", "ignore").decode(),  # we remove non-ascii chars
                last_update=n.last_update,
                address=address,
                color=n.color)

        for e in raw_graph.edges:
            policy1 = {
                'time_lock_delta': e.node1_policy.time_lock_delta,
                'fee_base_msat': e.node1_policy.fee_base_msat,
                'fee_rate_milli_msat': e.node1_policy.fee_rate_milli_msat,
                'last_update': e.node1_policy.last_update,
                'disabled': e.node1_policy.disabled,
                'min_htlc': e.node1_policy.min_htlc,
                'max_htlc_msat': e.node1_policy.max_htlc_msat
            }
            policy2 = {
                'time_lock_delta': e.node2_policy.time_lock_delta,
                'fee_base_msat': e.node2_policy.fee_base_msat,
                'fee_rate_milli_msat': e.node2_policy.fee_rate_milli_msat,
                'last_update': e.node2_policy.last_update,
                'disabled': e.node2_policy.disabled,
                'min_htlc': e.node2_policy.min_htlc,
                'max_htlc_msat': e.node2_policy.max_htlc_msat
            }
            # create a dictionary for channel_id lookups
            self.edges[e.channel_id] = {
                'node1_pub': e.node1_pub,
                'node2_pub': e.node2_pub,
                'capacity': e.capacity,
                'last_update': e.last_update,
                'channel_id': e.channel_id,
                'chan_point': e.chan_point,
                'policies': {
                    e.node1_pub > e.node2_pub: policy1,
                    e.node2_pub > e.node1_pub: policy2
                }
            }

            # add vertices to network graph for edge-based lookups
            self.graph.add_edge(
                e.node1_pub,
                e.node2_pub,
                channel_id=e.channel_id,
                last_update=e.last_update,
                capacity=e.capacity,
                fees={
                    e.node1_pub > e.node2_pub: policy1,
                    e.node2_pub > e.node1_pub: policy2,
                })

    def number_channels(self, node_pub_key):
        """
        Determines the degree of a given node.

        :param node_pub_key: str
        :return: int
        """
        try:
            number_of_channels = self.graph.degree[node_pub_key] / 2
        except KeyError:
            number_of_channels = 0
        return number_of_channels

    def node_capacity(self, node_pub_key):
        """
        Calculates the total capacity of a node in satoshi.

        :param node_pub_key: str
        :return: int
        """
        total_capacity = 0
        edges = self.graph.edges(node_pub_key, data=True)
        for e in edges:
            total_capacity += e[2]['capacity']
        return total_capacity

    def node_alias(self, node_pub_key):
        """
        Wrapper to get the alias of a node given its public key.

        :param node_pub_key:
        :return: alias string
        """
        try:
            return self.graph.nodes[node_pub_key]['alias']
        except KeyError:
            return 'unknown alias'

    def node_address(self, node_pub_key):
        """
        Returns the IP/onion addresses of a node.

        :param node_pub_key:
        :return: list
        """
        return self.graph.nodes[node_pub_key]['address']

    def node_age(self, node_pub_key):
        """
        Determine the approximate node's age by its oldest channel.

        :param node_pub_key: str
        :return: float, days
        """

        # find all channels of nodes
        node_edges = self.graph.edges(node_pub_key, data=True)

        # collect all channel's ages in terms of blockheights
        channel_ages = []
        for e in node_edges:
            channel_id = e[2]['channel_id']
            height, index, output = convert_channel_id_to_short_channel_id(channel_id)
            channel_age = self.node.blockheight - height
            channel_ages.append(channel_age)

        # determine oldest channel's age
        node_age = max(channel_ages)

        # convert to days from actual blockheight
        node_age_days = float(node_age) * 10 / (60 * 24)
        return node_age_days

    def neighbors(self, node_pub_key):
        """
        Finds all the node pub keys of nearest neighbor nodes.

        :param node_pub_key:  str
        :return: node_pub_key: str
        """
        neighbors = nx.neighbors(self.graph, node_pub_key)
        for n in neighbors:
            yield n

    def second_neighbors(self, node_pub_key):
        """
        Finds all the node pub keys of second nearest neighbor nodes (multiple appearances allowed).

        :param node_pub_key: str
        :return: node_pub_key: str
        """
        for neighbor_list in [self.graph.neighbors(n) for n in self.graph.neighbors(node_pub_key)]:
            for n in neighbor_list:
                yield n

    def nodes_in_neighborhood_of_nodes(self, nodes, blacklist_nodes, nnodes=100):
        """
        Takes a list of nodes and finds the neighbors with most connections to the nodes.

        :param nodes: list
        :param blacklist_nodes: list of node_pub_keys to be excluded from counting
        :param nnodes: int, limit for the number of nodes returned
        :return: list of tuples, (str pub_key, int number of neighbors)
        """
        nodes = set(nodes)
        # eliminate blacklisted nodes
        nodes = nodes.difference(blacklist_nodes)
        neighboring_nodes = []
        for general_node in self.graph.nodes:
            neighbors_general_node = set(self.neighbors(general_node))
            intersection_with_nodes = nodes.intersection(neighbors_general_node)
            number_of_connection_with_nodes = len(intersection_with_nodes)
            neighboring_nodes.append((general_node, number_of_connection_with_nodes))

        sorted_neighboring_nodes = sorted(neighboring_nodes, key=lambda x: x[1], reverse=True)
        return sorted_neighboring_nodes[:nnodes]


if __name__ == '__main__':
    import logging.config
    logging.config.dictConfig(settings.logger_config)

    from lndmanage.lib.node import LndNode
    nd = LndNode('')
    print(f"Graph size: {nd.network.graph.size()}")
    print(f"Number of channels: {len(nd.network.edges.keys())}")



from unittest import TestCase
import logging
import sys

from lndmanage.lib.fee_setting import delta_demand, delta_min, optimization_parameters

logger = logging.getLogger()
logger.level = logging.DEBUG
logger.addHandler(logging.StreamHandler(sys.stdout))


class TestFeeSetter(TestCase):
    def test_delta_min(self):
        cap = 2000000
        # maximal upward adjustment for empty local balance
        self.assertAlmostEqual(
            1 + optimization_parameters["delta_min_up"],
            delta_min(optimization_parameters, local_balance=0, capacity=cap),
        )
        # no adjustment if local balance is balance reserve
        self.assertAlmostEqual(
            1,
            delta_min(
                optimization_parameters,
                local_balance=optimization_parameters["local_balance_reserve"],
                capacity=cap,
            ),
        )
        # maximal downward adjustment for full local balance
        self.assertAlmostEqual(
            1 - optimization_parameters["delta_min_dn"],
            delta_min(optimization_parameters, local_balance=cap, capacity=cap),
        )

    def test_factor_demand_fee_rate(self):
        cap = 2000000
        interval_days = 7
        # maximal downward adjustment for full local balance and no demand
        self.assertAlmostEqual(
            1 - optimization_parameters["delta_min_dn"],
            delta_demand(
                optimization_parameters,
                time_interval=interval_days,
                amount_out=0,
                local_balance=cap,
                capacity=cap,
            ),
            places=6,
        )

        # maximal upward adjustment in the case of empty local balance and no demand
        self.assertAlmostEqual(
            1 + optimization_parameters["delta_min_up"],
            delta_demand(
                optimization_parameters,
                time_interval=interval_days,
                amount_out=0,
                local_balance=0,
                capacity=cap,
            ),
            places=6,
        )

        # optimal amount: no change
        self.assertAlmostEqual(
            1,
            delta_demand(
                optimization_parameters,
                time_interval=interval_days,
                amount_out=optimization_parameters["r_t"] * interval_days,
                local_balance=cap,
                capacity=cap,
            ),
            places=6,
        )

        # maximal demand: highest change
        self.assertAlmostEqual(
            optimization_parameters["delta_max"],
            delta_demand(
                optimization_parameters,
                time_interval=interval_days,
                amount_out=1000000,
                local_balance=cap,
                capacity=cap,
            ),
            places=6,
        )
